;***************************************************************************	

; IRdongle Firmware VERSION 2.1

; Features:
;	- IR Interpreter fr alle (!) Fernbedienungen durch Periodenanpassung
;	- Software UART fr PC-Kommunikation mit bis zu 115200 Baud

; Hardware:
;	- AT90S1200 @4 MHz
;	- IR-Dekoder TSOPxxxx, SFH51xx, etc

; Pinbelegung:
;	- PD0,1 = Rx,Tx
;	- PD3 = LED (Debugzwecke, spter vielleicht zum Senden)
;	- PD2 = IR-Signal


;***************************************************************************	
;	Includes

.include "1200def.inc"


;***************************************************************************	
;	Pinassignments

.equ	RxD	=0	; HW-UART
.equ	TxD	=1	; HW-UART
.equ	IRin	=2
.equ	LEDs	=PORTD
.equ	LED	=6

.equ	STARTBC	=0x01	; Bitschieberstartwert


;***************************************************************************	
;	Register assignments

.def	bitcnt	=R29			;bit counter
.def	temp	=R28			;temporary storage register

.def	Txbyte	=R27			;Data to be transmitted
.def	empfcnt	=R26

.def	flags	=R31

.def	t0temp	=R23			;T0 Temporary Register
.def	siglen	=R23			;Signal length of actual data bit
.def	siglo	=R24			;Interval lower bound
.def	sighi	=R25			;Interval upper bound

.def	WAIT	=R24			;Fr wait_us: Dauer in wait*100s
.def	WAIT2	=R25			;wait_us interner Zhler

.def	StartON	=R0
.def	StartOFF=R1
.def	ON0	=R2
.def	ON1	=R3
.def	OFF0	=R4
.def	OFF1	=R5
.def	Anzbits	=R6
.def	SplitPos=R7
.def	Splitlen=R8

;***************************************************************************	
;	Programmblock, Sprungtabelle

.cseg

.ORG	0				; *** reset vector 
	rjmp	 _reset

.ORG	INT0addr
	rjmp	 extint0

.ORG	OVF0addr
	rjmp	 t0int

;.ORG	URXCaddr			; *** Uart Interupt vector
;	rjmp	 fifoin

;.ORG	UTXCaddr
;	rjmp	 fifout


;***************************************************************************	
;	Interrupt EXT INT 0
extint0:	; ndert: R0, FLEXT0

	in	r0,	SREG	; SAVE SREG
	ori	flags,	1	; Flag Bit 0 setzen
	in	siglen,	TCNT0	; Timerstand zur Auswertung sichern
	cbi	LEDs,	LED	; LED an
	out	SREG,	r0	; RESTORE SREG
	reti


;***************************************************************************	
;	Interrupt TIMER 0
t0int:		; ndert: R0, SIGLEN, FLT0

	in	r0,	SREG	; SAVE SREG
	sbi	LEDs,	LED	; LED aus
	ori	flags,	2	; Flag Bit 1 setzen
	in	t0temp,	GIMSK	; Falls Empfangsmodus aktiv (Ext Int aktiviert)
	sbrc	t0temp,	6
	rjmp	t0exit		; verlasse Interrupt-Routine
	
 	ldi	t0temp,	0x97
 	out	TCNT0,	t0temp	;//set count value
 	clr	t0temp
	out	PORTB,	t0temp	;//LEDs an
	ldi	t0temp,	5	;Wartet 6us
  wloop:	
	subi	t0temp,	1
	cpi	t0temp,	0
	brne	wloop
	ldi	t0temp,	0xFF
	out	PORTB,	t0temp	;//LEDs aus
	reti

  t0exit:
  	out	SREG,	r0	; RESTORE SREG
	reti


;***************************************************************************	
;	Interrupt UART RX COMPLETE
;fifoin:	reti


;***************************************************************************	
;	Interrupt UART TX COMPLETE
;fifout:	reti


;***************************************************************************	
;	Programmblock, Startup

_reset:
;	ldi	temp,	RAMEND	; set up stack
;	out	SPL,	temp

	; PORTD Direction Setup
	ldi	temp,	0x42	; Bit 6 und 1 sind Ausgnge
	out	DDRD,	temp
	; PORTB Direction Setup
	ldi	temp,	0x01	; LED Array
	out	DDRB,	temp
	out	PORTB,	temp	; LED aus
	; EXTINT Setup
	ldi	temp,	0x40	; Ext. Int0 aktiviert
	out	GIMSK,	temp
	; TIMER0 Setup
	ldi	temp,	0x02	; T0 Int aktiviert
	out	TIMSK,	temp
	clr	temp		; Vorteiler des Timers = 1/256 (ca. 16ms), Timer Stopp
	out	TCCR0,	temp
	out	TCNT0,	temp
	; MCU Setup
	ldi	temp,	0x02	; Ext. Int0 reagiert nur auf fallende Flanken
	out	MCUCR,	temp
	; WATCHDOG Setup
	;wdr			; WDT Reset
	;ldi	temp,	0x05	; Setzt Resetwartezeit auf 1/2 Sekunde
	;out	WDTCR,	temp	
	
	sei			; Startet Interruptsystem

	; Startzustand herstellen
	
	cbi	LEDs,	LED	; LED an
	ldi	txbyte,	'I'
	rcall	putchar
	ldi	txbyte,	'R'
	rcall	putchar
	ldi	txbyte,	'-'
	rcall	putchar
	ldi	txbyte,	'D'
	rcall	putchar
	ldi	txbyte,	'o'
	rcall	putchar
	ldi	txbyte,	'n'
	rcall	putchar
	ldi	txbyte,	'g'
	rcall	putchar
	ldi	txbyte,	'l'
	rcall	putchar
	ldi	txbyte,	'e'
	rcall	putchar
	ldi	txbyte,	' '
	rcall	putchar
	ldi	txbyte,	'v'
	rcall	putchar
	ldi	txbyte,	'2'
	rcall	putchar
	ldi	txbyte,	'.'
	rcall	putchar
	ldi	txbyte,	'1'
	rcall	putchar
	ldi	txbyte,	13
	rcall	putchar
	sbi	LEDs,	LED	; LED aus
	
pdn:
	clr	flags
pdn2:	ldi	ZL,	1	; Startzustand: Z zeigt auf r1
	clr	empfcnt		; Bitzhler auf 0
	ldi	bitcnt,	STARTBC	; Bitschieber belegen
slp:	;ldi	temp,	0x22	; Power down!
	;out	MCUCR,	temp
  	;sleep

;***************************************************************************	
;	Hauptschleife MAIN

main:	sbrs	flags,	0	; Falls externer Int.: Einspeichern der Daten
	rjmp	nm
	sbrc	flags,	7	; Debugmodus an = Flag 7 an
	rjmp	debugmode

	;*********************************************************
		; Block of EXTINT0
		; ndert: FLEXT0, EMPFCNT, BITCNT, TEMP, ZL(r30)
		; Ruft auf: -
		
		cbr	flags,	1	; Flag lschen (Bit 0)
		
		cpi	empfcnt,2	; Empf.zhler auswerten
		brlo	retempfcnt2	; Wenn E.z.<2 dann macht er hier gar nichts
		brne	eqempfcnt	; Wenn E.z.=2 dann werden SIGLO und SIGHI berechnet

		mov	siglo,	siglen
		mov	temp,	siglen
		lsr	temp
		lsr	temp
		sub	siglo,	temp	; In SIGLO steht jetzt SIGLEN - SIGLEN/4
		mov	sighi,	siglen
		add	sighi,	temp	; In SIGHI ist jetzt SIGLEN + SIGLEN/4
	  retempfcnt2:
	  	inc	empfcnt
	  	rjmp	retnm

	  aexit2:
		sbrc	flags,	2	; Falls das Polbit bereits gesetzt (Bit 2)
		rjmp	aexit		; springe zurck
	
		ori	flags,	4	; Setze Flag um zu zeigen dass Polbit gesetzt wurde
		cp	SIGLEN,	SIGLO	; Falls neues Signal lnger ist setze T
		brlo	aexit
		bset	6		; Setze T Flag
		rjmp	aexit
	
	  eqempfcnt:
		; Auswerteroutine: Wenn SIGLEN zwischen SIGLO und SIGHI liegt, dann r[Z] OR bitcnt
		cp	SIGLEN,	SIGLO
		brlo	aexit2
		cp	SIGLEN,	SIGHI
		brsh	aexit2

	  writebit:	
		ld	temp,	Z	; Ldt das aktuelle Byte von r[Z]
		or	temp,	bitcnt	; Verodern mit Bitschieber
		st	Z,	temp	; Zurckschreiben
		
	  aexit:; bitcnt schieben und falls ins Carry geschoben, Z++ und bitcount auf Startbelegung
	  	rcall	bitschieben
;	  	lsl	bitcnt		; Bit schieben
;	  	brsh	retempfcnt	; Carry=0: berspringen
	  	
;	  	cpi	ZL,	21	; Schutzfunktion fr die Arbeitsregister
;	  	breq	retempfcnt
;	  	inc	ZL		; Z++
;	  	ldi	bitcnt,	STARTBC	; bitcnt mit Startwert belegen
		
;	  retempfcnt:
;		inc	empfcnt		; Empfangszhler um 1 erhhen
		
	  retnm:clr	temp
		out	TCCR0,	temp	; Timer0 stoppen
		out	TCNT0,	temp	; Timer0 zurcksetzen
		ldi	temp,	0x04
		out	TCCR0,	temp	; Timer0 starten
		
		
  nm:	sbrs	flags,	1	; Falls Timerberlauf: Senden der empf. Daten und Wiederherst. d. Startzustands
	rjmp	eom
	sbrc	flags,	7	; Debugmodus an = Flag 7 an
	rjmp	debugmode2

	;*********************************************************
	
		; Block of T0INT
		; ndert: FLT0, TXBYTE, BITCNT, TEMP, ZL(r30)
		; Ruft auf: -

		cbr	flags,	2	; Lsche Flag (Bit 1)
		clr	temp
		out	TCCR0,	temp	; Stoppe Timer0
		
		mov	empfcnt,ZL	; Speichert den aktuellen Z-Stand als Schleifenabbruchbedingung
		inc	empfcnt
		ldi	ZL,	1	; Startet immer bei r1
	  sl1:	
		ld	temp,	Z	; Ldt den Wert von r[Z]
		brtc	goon		; Falls T = 0 mache weiter
		com	temp		; ansonsten invertiere das zu sendende Byte
	  goon:
	  	mov	txbyte,	temp	; Bereitet Senden vor
		swap	txbyte		; Tauscht Nibbles
		rcall	sendnibble	; Sendet den Hexwert der oberen 4 Bits als HEX
  		
		mov	txbyte,	temp	; Ldt den Wert
		rcall	sendnibble	; Sendet den Hexwert der unteren 4 Bits als HEX

		clr	temp		; Lscht das Byte an r[Z]
		st	Z,	temp  		
		inc	ZL		; Z++
		cp	ZL,	empfcnt	; Solange Z kleiner als der Empfangszhler ist, wird gesendet
		brlo	sl1		; Sprung zum Senden
		
		ldi	txbyte,	13	; Abschlusszeichen
		rcall	putchar		; senden
		
		rjmp	pdn2
		
  eom:	sbis	PIND,	RxD	; Leitung geht auf 1 -> Einleseroutine wird gestartet
  	rjmp	slp
	;*********************************************************
	
		; Block of RXCHAR
		; ndert: TEMP, TXBYTE, BITCNT, EMPFCNT, R0..R22, Z, WAIT, WAIT2
		; Ruft auf: GETNIBBLE
		
		ldi	ZL,	1	; Zhler bei r1 starten
		clr	empfcnt		; Bitzhler lschen
		out	GIMSK,	empfcnt	; Ext int abschalten
	 	out	TCCR0,	empfcnt	;//stop timer
		
	  lrcv:
		clr	temp		; Temp lschen um #13 zu erkennen
		rcall	getnibble	; Oberes Nibble einlesen (0..F)
		cpi	temp,	13	; Falls Temp = 13 dann wurde kein Nibble sondern #13 empf
		breq	recvexit	; und wir verlassen die Einleseroutine
		cpi	temp,	10	; Debug an
		brne	ndebugan
		
		; Code fr DEBUG mode an
	  debugan:
		ori	flags,	0x80	; Debug flag setzen
		ldi	txbyte,	13
		rcall	putchar
	  dsn:	ldi	temp,	0x40	; Ext. Int0 aktiviert
		out	GIMSK,	temp
		rjmp	pdn2		; Neustart
		
	  ndebugan:	
		cpi	temp,	11	; Debug aus
		brne	ndebugaus
		
		; Code fr DEBUG mode aus
	  debugaus:
		cbr	flags,	0x80	; Debug flag lschen
		rjmp	dsn
		
	  ndebugaus:
		mov	wait,	txbyte	; Empf Nibble in die oberen 4 Bits bringen
		swap	wait
		rcall	getnibble	; Unteres Nibble lesen
		cpi	temp,	13	; Falls Temp = 13 dann wurde kein Nibble sondern #13 empf
		breq	recvexit	; und wir verlassen die Einleseroutine
		cpi	temp,	10	; Debug an
		breq	debugan
		cpi	temp,	11	; Debug aus
		breq	debugaus
		or	txbyte,	wait	; Odern
		st	Z,	txbyte	; An r[Z] speichern
		inc	ZL		; Z++
		cpi	ZL,	23	; Falls Z < 23
		brlo	lrcv		; Noch einen empfangen
		
	  recvexit:
		; Hier musst du den Sendealgorithmus entwickeln!
		; Erste Version ohne Split
 		ldi	temp,	0x97
	 	out	TCNT0,	temp	;//set count value
		ldi	ZL,	1	; Zhler bei r1 starten
		out	TCCR0,	ZL	; Timer 36kHz

		ldi	bitcnt,	1	; Bitschieber
		ldi	temp,	0xFF	; LED an
		out	DDRB,	temp
		mov	WAIT,	StartOn	; wait_us fr r0
		rcall	wait_us
		inc	temp		; LED aus
		out	DDRB,	temp
		mov	WAIT,	StartOff; wait_us fr r1
		rcall	wait_us

	  sloop:dec	temp		; LED an
		out	DDRB,	temp
		
		ld	txbyte,	Z	; Byte laden
		and	txbyte,	bitcnt	; Mit Bitschieber maskieren

;		cp	empfcnt,Splitpos
;		breq	spos1

		mov	WAIT,	ON0
		cpi	txbyte,	0	; Falls Bit = 0
		breq	wr
		mov	WAIT,	ON1
	  wr:	rcall	wait_us
		
		rcall	bitschieben
		cp	empfcnt,Anzbits	; Abbruchbedingung Anzahl Bits
		breq	exsenden
		
	  	inc	temp		; LED aus
		out	DDRB,	temp
		
		ld	txbyte,	Z	; Byte laden
		and	txbyte,	bitcnt	; Mit Bitschieber maskieren
		
	  	mov	WAIT,	OFF0	; Warte gewnschte Zeit
		cpi	txbyte,	0	; Falls Bit = 0
		breq	wr1
		mov	WAIT,	OFF1
	  wr1:	rcall	wait_us
		
		rcall	bitschieben
		cp	empfcnt,Anzbits	; Abbruchbedingung Anzahl Bits
		brne	sloop
		
	  exsenden:	
		; Startzustand wiederherstellen
		clr	temp		; LED aus
		st	Z,	temp	; Byte lschen
		out	DDRB,	temp	
		out	TCCR0,	temp	; Timer0 aus
		out	TCNT0,	temp
		ldi	temp,	0x40	; Ext Int wieder aktivieren
		out	GIMSK,	temp
		ldi	WAIT,	10	; Kleine Pause
		rjmp	pdn		; Zur Startbelegung springen
	

;***************************************************************************	
debugmode:	; ndert: TXBYTE, TEMP
		; Ruft auf: PUTCHAR

	mov	txbyte,	siglen	; Timerstand senden!
	rcall	putchar
	ldi	temp,	4
	out	TCCR0,	temp
	clr	temp		; Timer zurcksetzen!
	out	TCNT0,	temp
	rjmp	slp


;***************************************************************************	
debugmode2:	; ndert: TEMP

	clr	temp		; Timer stoppen!
	out	TCCR0,	temp
	rjmp	slp
	
	
;***************************************************************************	
bitschieben:	; ndert: BITCNT, ZL, EMPFCNT

  	lsl	bitcnt		; Bit schieben
  	brsh	retempfcnt	; Carry=0: berspringen
	  	
  	cpi	ZL,	21	; Schutzfunktion fr die Arbeitsregister
  	breq	retempfcnt
  	inc	ZL		; Z++
  	ldi	bitcnt,	STARTBC	; bitcnt mit Startwert belegen

  retempfcnt:
	inc	empfcnt		; Empfangszhler um 1 erhhen
	ret


;***************************************************************************	
sendnibble:	; ndert TXBYTE, BITCNT
		; Ruft auf: PUTCHAR
		
	andi	txbyte,	0x0F
	ldi	bitcnt,	48
	add	txbyte,	bitcnt
	cpi	txbyte,	58
	brlo	sn1
	
	ldi	bitcnt,	7
	add	txbyte,	bitcnt
  sn1:	rcall	putchar
  	ret


;***************************************************************************	
getnibble:	; ndert: TXBYTE
		; Ruft auf: getchar
		
	rcall	getchar
	mov	temp,	txbyte
	cpi	txbyte,	48	; Abbruch bei Steuerzeichen
	brlo	txexit
	subi	txbyte,	48
	cpi	txbyte,	10
	brlo	gn1
	
	subi	txbyte,	7
  gn1:
  	andi	txbyte,	0x0F
  txexit:
  	ret
	

;***************************************************************************	
;loop:
;	in	temp,	PIND
;	andi	temp,	0x04
;	cpi	temp,	0x04
;	breq	l2
	
;	clr	temp
;	out	PORTD,	temp
;	rjmp	loop
 ; l2:
;  	ldi	temp,	0x02
;  	out	PORTD,	temp
;	rjmp	loop
	
;***************************************************************************
;*
;* "putchar"
;*
;* This subroutine transmits the byte stored in the "Txbyte" register
;* The number of stop bits used is set with the sb constant
;*
;* Number of words	:14 including return
;* Number of cycles	:Depens on bit rate
;* Low registers used	:None
;* High registers used	:2 (bitcnt,Txbyte)
;* Pointers used	:None
;*
;***************************************************************************
.equ		sb	=1		;Number of stop bits (1, 2, ...)

putchar:	ldi	bitcnt,9+sb	;1+8+sb (sb is # of stop bits)
		com	Txbyte		;Inverte everything
		sec			;Start bit

putchar0:	brcc	putchar1	;If carry set
		sbi	PORTD,TxD	;    send a '0'
		rjmp	putchar2	;else	

putchar1:	cbi	PORTD,TxD	;    send a '1'
		nop

putchar2:	rcall UART_delay	;One bit delay
		rcall UART_delay

		lsr	Txbyte		;Get next bit
		dec	bitcnt		;If not all bit sent
		brne	putchar0	;   send next
					;else
		ret			;   return


;***************************************************************************
;*
;* "UART_delay"
;*
;* This delay subroutine generates the required delay between the bits when
;* transmitting and receiving bytes. The total execution time is set by the
;* constant "b":
;*
;*	3b + 7 cycles (including rcall and ret)
;*
;* Number of words	:4 including return
;* Low registers used	:None
;* High registers used	:1 (temp)
;* Pointers used	:None
;*
;***************************************************************************
; Some b values: 	(See also table in Appnote documentation)
;
; 1 MHz crystal:
;   9600 bps - b=14
;  19200 bps - b=5
;  28800 bps - b=2
;
; 2 MHz crystal:
;  19200 bps - b=14
;  28800 bps - b=8
;  57600 bps - b=2

; 4 MHz crystal:
;  19200 bps - b=31
;  28800 bps - b=19
;  57600 bps - b=8
; 115200 bps - b=2

.equ	b	=2	;115200 bps @ 4 MHz crystal


UART_delay:	ldi	wait,b
UART_delay1:	dec	wait
		brne	UART_delay1

		ret

;***************************************************************************
;*
;* "getchar"
;*
;* This subroutine receives one byte and returns it in the "Rxbyte" register
;*
;* Number of words	:14 including return
;* Number of cycles	:Depens on when data arrives
;* Low registers used	:None
;* High registers used	:2 (bitcnt,Rxbyte)
;* Pointers used	:None
;*
;***************************************************************************

getchar:	ldi 	bitcnt,9	;8 data bit + 1 stop bit

getchar1:	sbis 	PIND,RxD	;Wait for start bit
		rjmp 	getchar1

		rcall UART_delay	;0.5 bit delay

getchar2:	rcall UART_delay	;1 bit delay
		rcall UART_delay		

		clc			;clear carry
		sbis 	PIND,RxD	;if RX pin high
		sec			;

		dec 	bitcnt		;If bit is stop bit
		breq 	getchar3	;   return
					;else
		ror 	Txbyte		;   shift bit into Rxbyte
		rjmp 	getchar2	;   go get next

getchar3:	ret


;***************************************************************************	
wait_us:	; Wartet ca. 100*WAIT s @ 4 MHz
		; ndert: WAIT, WAIT2
		
	ldi	WAIT2,	0x83
  __long_delay0:
	subi	WAIT2,	1	;2 cycles
	brne	__long_delay0 	;2 cycles
	subi	WAIT,	1
	brne	wait_us 	;2 cycles
	ret

; Senden von IR Codes:
; Header (7-9):	Startsignal an/aus Lnge (2)
;		Signal 0 an/aus Lnge (2)
;		Signal 1 an/aus Lnge (2)
;		Signal Bitanzahl und Split an/aus Bit (1)
;		Falls Split = an Position des Split und Lnge (2)
; Daten:	Bitfolge entsprechend Anzahl (1-11/13)
; bertragung vom PC: HEX-Codes, Return nach Header und nach Bits
; Bits werden von unten her gefllt, LSB first (ROR zum Senden)


